#define	VERSION	"Poppassd v1.1 (NT)"

/*	
	     Program: Poppassd
Operating System: Window NT
         Version: 1.1

This program will only work for NT machines.

	also the following lib's have to be included
		

 * poppassd_nt.c
 *
 * A Eudora and NUPOP change password server.
 *
 * This is an NT version of poppassd.c designed by
 *
 * John Norstad
 * Academic Computing and Network Services
 * Northwestern University
 * j-norstad@nwu.edu
 *
 * Based on earlier versions by Roy Smith <roy@nyu.edu> and Daniel
 * L. Leavitt <dll.mitre.org>.
 *  
 * Doesn't actually change any passwords itself.  It simply requests
 * the operating system to preform the password change.
 *
 * Note that unencrypted passwords are transmitted over the network.  If
 * this bothers you, think hard about whether you want to implement the
 * password changing feature.  
 *
 * NOTE: In addition to the security issue outlined in the above paragraph,
 * you should be aware that this program is going to be run as root by
 * ordinary users and it mucks around with the password file.  This should
 * set alarms off in your head.  I think I've devised a pretty foolproof
 * way to ensure that security is maintained, but I'm no security expert and
 * you would be a fool to install this without first reading the code and
 * ensuring yourself that what I consider safe is good enough for you.  If
 * something goes wrong, it's your fault, not mine.
 *
 * The front-end code (which talks to the client) is directly 
 * descended from Leavitt's original version.  The back-end pseudo-tty stuff 
 * (which talks to /bin/password) is directly descended from Smith's
 * version, with changes for SunOS and HP-UX by Norstad (with help from
 * sample code in "Advanced Programming in the UNIX Environment"
 * by W. Richard Stevens). The code to report error messages
 * back to the client in the final 500 response, and a new version of the
 * code to find the next free pty, is by Norstad.
 *        

Installing poppassd_nt.c
 *
 * First you will need to compile the source code. Make sure that you
 * include the libarary 
 *						wsock32.lib
 *						netapi32.lib
 * as with this libarary the compiling will not work.
 *
 * Only admin's can install poppassd.
 *
 * The compiled version is then copied to the directory.
 *
 *				'\winnt\system32'
 *
 * You will then have to set the program as a service, this is 
 * done my running the 'addsvr.exe' that is provided see below.
 *
 *		addsvc -add poppassd poppassd.exe "poppassd -- Changing Passwords"
 *
 * To remove the poppassd from the service use the following line
 *
 *		addsvc -del poppassd
 *
 * Once setup goto the service's 
 *		(ie start -> settings -> control pannel -> services)
 *
 * select the open labeled "poppassd -- Changing Passwords"
 *
 * Click on the startup button, and change the 'Log On As'
 * to 'This Account' and select an account that has access
 * rights to change passwords, also enter the password.
 *
*/
/* Steve Dorner's description of the simple protocol:
 *
 * The server's responses should be like an FTP server's responses; 
 * 1xx for in progress, 2xx for success, 3xx for more information
 * needed, 4xx for temporary failure, and 5xx for permanent failure.  
 * Putting it all together, here's a sample conversation:
 *
 *   S: 200 machine_name popassd v1.4 hello, who are you?\r\n
 *   E: user yourloginname\r\n
 *   S: 200 your password please.
 *   E: pass yourcurrentpassword\r\n
 *   S: 200 your new password please.\r\n
 *   E: newpass yournewpassword\r\n
 *   200 Password changed, thank-you.\r\n
 *   E: quit\r\n
 *   S: 200 Bye-bye\r\n
 *   S: <closes connection>
 *   E: <closes connection>
 */



#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>
#include <string.h>
#include <ctype.h>


#include <winsock.h>
#define socket_close closesocket
#define socket_ioctl ioctlsocket
#define socket_read(s,b,l) recv(s,b,l,0)
#define socket_write(s,b,l) send(s,b,l,0)

#include <windows.h>
#include <lmcons.h>
#include <lmaccess.h>
#include <lmerr.h>
#include <lmapibuf.h>


char *error_string = NULL;


FILE *log_file = NULL;

void set_log(char *file)
{
	log_file = fopen(file,"a");
}

void display(char *string,...) 
{
	va_list arg_ptr;
  	char data[1024];

	va_start(arg_ptr, string);
	vsprintf(data, string, arg_ptr);

	if (log_file) {
		fprintf(log_file,data);	
		fflush(log_file);
	}
	printf(data);
}

/************************************************************************/
/************************************************************************/
/*	The below functions are to make the Program a Service */

#define SVC_NAME "poppassd"
int finished = 0;


#include <winnt.h>

char *svc_getargs(char *svcname);
SERVICE_STATUS_HANDLE ssh;
VOID svc_handler(DWORD msg);
void suck_setexit(void);

void svc_report(int status)
{
	static int cp;
	static int stop_pending;
	static SERVICE_STATUS ss;
	
	ss.dwServiceType = SERVICE_WIN32_OWN_PROCESS;
	ss.dwCurrentState = status;	 
	ss.dwControlsAccepted = SERVICE_ACCEPT_SHUTDOWN | SERVICE_ACCEPT_STOP;
	ss.dwWin32ExitCode = NO_ERROR;
	ss.dwServiceSpecificExitCode = 0;
	ss.dwWaitHint = 0;
	ss.dwCheckPoint = 0;
	
	if ((status==SERVICE_STOP_PENDING) || (status==SERVICE_START_PENDING)) {
		ss.dwWaitHint = 30;
		ss.dwCheckPoint = cp++;
	}
	if ((status==SERVICE_STOP_PENDING)) {
		ss.dwWaitHint = 10;
		ss.dwCheckPoint = cp++;
	}

	if (ssh!=0) SetServiceStatus(ssh,&ss);
}

VOID svc_handler(DWORD msg)
{
	static int cp;
	static int stop_pending;
	int mde = SERVICE_RUNNING;	

	switch (msg) {
		case SERVICE_CONTROL_PAUSE:
		case SERVICE_CONTROL_CONTINUE:
		case SERVICE_CONTROL_INTERROGATE:
			break;
		case SERVICE_CONTROL_SHUTDOWN:
		case SERVICE_CONTROL_STOP:
			mde = SERVICE_STOP_PENDING;
			break;
	}
	svc_report(mde);
	if (mde==SERVICE_STOP_PENDING) {
		display("Inform Program of Shutdown");
		finished = 1;

	/* Shutdown Service NOW */
		display("Exiting Service");
		svc_report(SERVICE_STOP_PENDING);		

		display("Shutdown Service");
		svc_report(SERVICE_STOPPED);
	}
}


char xname[200];
int ServiceMain(int xargc, char *xargv[]);

int main(int argc, char *argv[])
{
	static SERVICE_TABLE_ENTRY stable[3];
	int isslave = FALSE;
	
	if (argc>1) {
		ServiceMain(argc,argv);
		return TRUE;
	}

	if (argc>0) 
		strcpy(xname,argv[0]);
	stable[0].lpServiceName = SVC_NAME;
	stable[0].lpServiceProc = (LPSERVICE_MAIN_FUNCTION) ServiceMain;
	stable[1].lpServiceName = NULL;
	stable[1].lpServiceProc = NULL;

	if (!StartServiceCtrlDispatcher(stable)) {
		LPVOID lpMsgBuf;
		int error = GetLastError();			
		
		if (error == ERROR_INVALID_DATA)
			display("Failed to setup as service {INVALID_DATA}\n");
		else if (error == ERROR_INVALID_NAME)
			display("Failed to setup as service {INVALID_NAME}\n");
		else  if (error == ERROR_SERVICE_DOES_NOT_EXIST)
			display("Failed to setup as service {SERVICE_DOES_NOT_EXIST}\n");
		else
			display("Failed to setup as service {%d}\n",error);
		

		FormatMessage( 
			FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM,
			NULL,
			error,
			MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), // Default language
			(LPTSTR) &lpMsgBuf,
			0,
			NULL 
		);

		display(">>%s",lpMsgBuf);
		LocalFree( lpMsgBuf );

	 	display("Starting %s\n",VERSION);
		ServiceMain(argc,argv);
	}
	svc_report(SERVICE_STOPPED);
	return 0;
}


void svc_init(char *sname)
{
	int r;
	ssh = RegisterServiceCtrlHandler(sname,(LPHANDLER_FUNCTION)	svc_handler);

	if (ssh==0) {
	 	r = GetLastError();
		if (r==ERROR_INVALID_NAME) display("svc: Invalid name\n");
		else if (r==ERROR_SERVICE_DOES_NOT_EXIST) display("svc: Service doesn't exist\n");
		else display("Register service failed, error %d \n",r);
	}
	svc_report(SERVICE_START_PENDING);
	svc_report(SERVICE_RUNNING);
}

char *svc_getargs(char *svcname)
{
	char key[400];
	static char val[400];
	int vallen=300,r;
	int valtype;
	HKEY skey;
	DWORD reserved=0;
	REGSAM sam;

	sprintf(key,"System\\CurrentControlSet\\Services\\%s",svcname);
	/* r = RegOpenKey(HKEY_LOCAL_MACHINE,key,&skey);*/
	sam = KEY_READ;
	r = RegOpenKeyEx(HKEY_LOCAL_MACHINE,key,reserved,sam,&skey);
	if (r!=ERROR_SUCCESS) display("failure: RegOpenKey %d(0x%02x)\n", r,GetLastError());
	r = RegQueryValueEx(skey,"DisplayName",NULL,&valtype,val,&vallen);
	if (r!=ERROR_SUCCESS) display("failure: RegQueryKey %d(0x%02x)\n", r,GetLastError());
	val[vallen] = 0;
	return val;
}



/************************************************************************/
/************************************************************************/
/*	The below functions are the socket functions */

/* Creating Connection */
int net_start()
{
	WORD wVersionRequested;	
	WSADATA wsaData;

 	wVersionRequested = MAKEWORD( 1, 1 );

	if ( !WSAStartup( wVersionRequested, &wsaData ) ) {
		if ( LOBYTE( wsaData.wVersion ) == 1 && HIBYTE( wsaData.wVersion ) == 1 ) 
			return TRUE;

		error_string = "No winsock found of correct version, trying to continue";
	}
	else
		error_string = "No winsock found of correct version";

	return FALSE;
}


SOCKET net_open(char *ip_address, int port)
{
	struct sockaddr_in xsin;
	struct hostent *hp = NULL;
	int st;

	SOCKET connection = INVALID_SOCKET;

	if (ip_address) {

	/* Double check inputs */
		if ( strlen( ip_address ) <= 3 ) {
			error_string = "Could not open net -- IP Address not set";
			return INVALID_SOCKET;
		}
		if ( port == 0 ) {
			error_string = "Could not open net -- Port not set";			
			return INVALID_SOCKET;
		}


		memset(&xsin,0,sizeof(xsin));
		xsin.sin_family = AF_INET;

		strlwr(ip_address); /* Seems to be required by NIS */

		if (isdigit(*ip_address)) {
			xsin.sin_addr.s_addr = inet_addr(ip_address);
			xsin.sin_family = AF_INET;
		} 
		else {
			xsin.sin_addr.s_addr = 0;
			hp = gethostbyname(ip_address);
			if (hp==NULL) {
				error_string = "gethostbyname failed.";
				return INVALID_SOCKET;
			}
			memcpy(&(xsin.sin_addr),hp->h_addr,  hp->h_length);
			xsin.sin_family = hp->h_addrtype;
		}

	/* Getting Socket */
		connection = socket(xsin.sin_family,SOCK_STREAM, 0);
		if (connection == INVALID_SOCKET) {
			error_string = "Failed to open socket correctly.";
			return INVALID_SOCKET;
		}
	
		xsin.sin_port = htons((unsigned short) port);

	/* Binding Socket */
		st = bind (connection,(struct sockaddr *)   &xsin, sizeof(xsin));
		if (st!=0) {
			error_string = "Bind to Socket Failed.";
			return INVALID_SOCKET;
		}
		
	/* Listen to socket */
		st = listen(connection,5);
		if (st) {
			error_string = "Listen to Socket Failed.";
			return INVALID_SOCKET;
		}

		return connection;
	}
	error_string = "Could not open net -- IP Address not set";
	return INVALID_SOCKET;
}

int net_send(SOCKET sock,char *string,...) 
{
	va_list arg_ptr;
  	char data[1024];

	va_start(arg_ptr, string);
	vsprintf(data, string, arg_ptr);

	return socket_write(sock,data,strlen(data));
}

char *net_read(SOCKET sock)
{
	static char data[1024];
	char *ptr = data;
	int x;

	data[0] = '\0';
/* Remove all '\r' and '\n' from input */
	*ptr = '\0';
	x = socket_read(sock,ptr,1);
	while (x && *ptr != '\n') {
		if (*ptr != '\r') ptr++;
		x = socket_read(sock,ptr,1);
	}
	*ptr = '\0';
	if (x)
		return strdup(data);
	
	return NULL;
}

/************************************************************************/
/************************************************************************/

LPWSTR convert_char(char *ptr) 
{
	LPWSTR wuser;
	wuser = malloc(2000);

	MultiByteToWideChar(
		CP_ACP,	// code page 
		0,		// character-type options 
		ptr,	// address of string to map 
		-1,		// number of characters in string 
		wuser,	// address of wide-character buffer 
		100 	// size of buffer 
		);

	return wuser;
}

/* Changing Password */
int change_password(char *user, char *pass, char *newpass) 
{
	int err;
	LPWSTR lpszUser;
    LPWSTR lpszPassword;
    LPWSTR lpszNewPassword;

	lpszUser = convert_char(user);
	lpszPassword = convert_char(pass);
	lpszNewPassword = convert_char(newpass);
    

	err =  NetUserChangePassword(
				NULL,lpszUser, lpszPassword, lpszNewPassword);
   
    switch ( err ) {
		case 0:
			return TRUE;
		case NERR_PasswordTooShort:
			error_string = "500 Password too short\r\n";
			break;
		case ERROR_ACCESS_DENIED:
			error_string = "500 Incorrect Username/Password\r\n";
			break;
		case NERR_InvalidComputer:
			error_string = "500 Incorrect Username/Password\r\n";
			break;
		case NERR_NotPrimary:
			error_string = "500 Incorrect Username/Password\r\n";
			break;
		case NERR_UserNotFound:
			error_string = "500 Incorrect Username/Password\r\n";
			break;
		default:
			error_string = "500 Incorrect Username/Password\r\n";
    }

	free(lpszUser);
    free(lpszPassword);
    free(lpszNewPassword);
	return FALSE;
}


void preform_connection(SOCKET sock)
{
	char *user=NULL, *pass=NULL, *newpass=NULL, *quit=NULL;	

/* Send Introduction */
/* Get input  *   E: user yourloginname\r\n */
	net_send(sock,"200 %s -- hello, who are you?\r\n",VERSION);
	user = net_read(sock);
	if (!user) return;

/* Ask for password */
	net_send(sock,"200 your password please.\r\n");
	pass = net_read(sock);
	if (!pass) {
		free(user);
		return;
	}

	if (!strnicmp(user,"user ",5) && !strnicmp(pass,"pass ",5)) {

	/* Ask for new Password */
		net_send(sock,"200 your new password please.\r\n");
		newpass = net_read(sock);

		if (newpass && !strnicmp(newpass,"newpass ",8)) {

			if (change_password(&user[5],&pass[5],&newpass[8])) {

				display("Changed '%s' password\n",&user[5]);
			/* Changed Password -- Wait for Quit */
				net_send(sock,"200 Password changed, thank-you.\r\n");
				quit = net_read(sock);

				if (quit) {
					if (!strnicmp(quit,"quit",4)) 
						net_send(sock,"200 Bye-bye\r\n");
					else
						net_send(sock,"500 Command Not Expected.\r\n");
				}
			}
			else {
				net_send(sock,error_string);
				display("Failed to change '%s' password\n",&user[5]);
			}
		}
		else
			net_send(sock,"500 Command Not Expected.\r\n");
	}
	else
		net_send(sock,"500 Command Not Expected.\r\n");


	free(user);
	free(pass);
	free(newpass);
	free(quit);
}

/************************************************************************/
/************************************************************************/
/* This is the Main Service Routine */

int ServiceMain(int xargc, char *xargv[])
{
	SOCKET connection, join;
	char bf[256];
	char *ip_address = "127.0.0.1", *str;
	FILE *stream;
	int port = 106;	

	if (xargc <= 1)
		svc_init(xargv[0]);

	set_log("c:\\poppassd.log");

/* Starting Network */	
	if (!net_start()) {
		display("Error -- %s\n", error_string);
		exit(1);
	}


/* Load IP address from file if needed */
	stream = fopen("ipaddress.dat","rb");
	if (stream) {
		
		if (fgets(bf,250,stream)) {
			str = bf + strlen(bf) - 1;
			while (*str == ' ' || *str == '\t' || *str == '\n' || *str == '\r')
				*str-- = '\0';

			ip_address = bf;
		}

		fclose(stream);
	}
	
/* Open Port on IP address */
	display("Connecting information.\n");			
	display("  IP Address = %s\n",ip_address);
	display("        Port = %d\n",port);

	connection = net_open(ip_address,port);
	if (connection == INVALID_SOCKET) {
		display("Error -- %s\n", error_string);
		exit(1);
	}

	
/* Sit and wait for connection */
	while (!finished) {

		join = accept(connection,NULL,0);

		if (join) {
			display("Accepted Channel {%d}\n",join);

			preform_connection(join);

			if (socket_close(join)<0)
				display("Channel close failed {%d}\n",connection);
			else
				display("Channel close successfully {%d}\n",connection);
		}
		else {
		/* Timeout */
			finished = 1;
		}
	}



/* Close connection */
	if (socket_close(connection)<0) { 
		display("Channel close failed {%d}\n",connection);
		exit(1);
	}
	display("Channel close successfully {%d}\n",connection);
	return 0;
}